@@ -21,16 +21,18 @@
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.
 */
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 #include <assert.h>
+#if 0
+#include <pthread.h>
+#endif
 #include <sys/types.h>
 #include "bgzf.h"
 
-#define UNUSED(x) ((void)(x))
+#define UNUSED(x) (void)(x)
 
 #ifdef _USE_KNETFILE
 #include "knetfile.h"
@@ -65,6 +67,7 @@ typedef FILE *_bgzf_file_t;
 #define BLOCK_HEADER_LENGTH 18
 #define BLOCK_FOOTER_LENGTH 8
 
+
 /* BGZF/GZIP header (speciallized from RFC 1952; little endian):
  +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  | 31|139|  8|  4|              0|  0|255|      6| 66| 67|      2|BLK_LEN|
@@ -105,9 +108,9 @@ static BGZF *bgzf_read_init()
 {
 	BGZF *fp;
 	fp = calloc(1, sizeof(BGZF));
-	fp->open_mode = 'r';
-	fp->uncompressed_block = malloc(BGZF_BLOCK_SIZE);
-	fp->compressed_block = malloc(BGZF_BLOCK_SIZE);
+	fp->is_write = 0;
+	fp->uncompressed_block = malloc(BGZF_MAX_BLOCK_SIZE);
+	fp->compressed_block = malloc(BGZF_MAX_BLOCK_SIZE);
 #ifdef BGZF_CACHE
 	fp->cache = kh_init(cache);
 #endif
@@ -118,9 +121,9 @@ static BGZF *bgzf_write_init(int compress_level) // compress_level==-1 for the d
 {
 	BGZF *fp;
 	fp = calloc(1, sizeof(BGZF));
-	fp->open_mode = 'w';
-	fp->uncompressed_block = malloc(BGZF_BLOCK_SIZE);
-	fp->compressed_block = malloc(BGZF_BLOCK_SIZE);
+	fp->is_write = 1;
+	fp->uncompressed_block = malloc(BGZF_MAX_BLOCK_SIZE);
+	fp->compressed_block = malloc(BGZF_MAX_BLOCK_SIZE);
 	fp->compress_level = compress_level < 0? Z_DEFAULT_COMPRESSION : compress_level; // Z_DEFAULT_COMPRESSION==-1
 	if (fp->compress_level > 9) fp->compress_level = Z_DEFAULT_COMPRESSION;
 	return fp;
@@ -128,7 +131,7 @@ static BGZF *bgzf_write_init(int compress_level) // compress_level==-1 for the d
 // get the compress level from the mode string
 // remove __restrict, as __ prefix is not portable.
 // static int mode2level(const char *__restrict mode)
-static int mode2level(const char *mode)
+static int mode2level(const char * mode)
 {
 	int i, compress_level = -1;
 	for (i = 0; mode[i]; ++i)
@@ -141,12 +144,13 @@ static int mode2level(const char *mode)
 BGZF *bgzf_open(const char *path, const char *mode)
 {
 	BGZF *fp = 0;
+	assert(compressBound(BGZF_BLOCK_SIZE) < BGZF_MAX_BLOCK_SIZE);
 	if (strchr(mode, 'r') || strchr(mode, 'R')) {
 		_bgzf_file_t fpr;
-#ifdef _WIN32        
+#ifdef _WIN32		
 		if ((fpr = _bgzf_open(path, "rb")) == 0) return 0;
 #else
-        if ((fpr = _bgzf_open(path, "r")) == 0) return 0;
+		if ((fpr = _bgzf_open(path, "r")) == 0) return 0;
 #endif
 		fp = bgzf_read_init();
 		fp->fp = fpr;
@@ -162,6 +166,7 @@ BGZF *bgzf_open(const char *path, const char *mode)
 BGZF *bgzf_dopen(int fd, const char *mode)
 {
 	BGZF *fp = 0;
+	assert(compressBound(BGZF_BLOCK_SIZE) < BGZF_MAX_BLOCK_SIZE);
 	if (strchr(mode, 'r') || strchr(mode, 'R')) {
 		_bgzf_file_t fpr;
 		if ((fpr = _bgzf_dopen(fd, "r")) == 0) return 0;
@@ -176,67 +181,42 @@ BGZF *bgzf_dopen(int fd, const char *mode)
 	return fp;
 }
 
+static int bgzf_compress(void *_dst, int *dlen, void *src, int slen, int level)
+{
+	uint32_t crc;
+	z_stream zs;
+	uint8_t *dst = (uint8_t*)_dst;
+
+	// compress the body
+	zs.zalloc = NULL; zs.zfree = NULL;
+	zs.next_in  = src;
+	zs.avail_in = slen;
+	zs.next_out = dst + BLOCK_HEADER_LENGTH;
+	zs.avail_out = *dlen - BLOCK_HEADER_LENGTH - BLOCK_FOOTER_LENGTH;
+	if (deflateInit2(&zs, level, Z_DEFLATED, -15, 8, Z_DEFAULT_STRATEGY) != Z_OK) return -1; // -15 to disable zlib header/footer
+	if (deflate(&zs, Z_FINISH) != Z_STREAM_END) return -1;
+	if (deflateEnd(&zs) != Z_OK) return -1;
+	*dlen = zs.total_out + BLOCK_HEADER_LENGTH + BLOCK_FOOTER_LENGTH;
+	// write the header
+	memcpy(dst, g_magic, BLOCK_HEADER_LENGTH); // the last two bytes are a place holder for the length of the block
+	packInt16(&dst[16], *dlen - 1); // write the compressed length; -1 to fit 2 bytes
+	// write the footer
+	crc = crc32(crc32(0L, NULL, 0L), src, slen);
+	packInt32((uint8_t*)&dst[*dlen - 8], crc);
+	packInt32((uint8_t*)&dst[*dlen - 4], slen);
+	return 0;
+}
+
 // Deflate the block in fp->uncompressed_block into fp->compressed_block. Also adds an extra field that stores the compressed block length.
 static int deflate_block(BGZF *fp, int block_length)
 {
-	uint8_t *buffer = fp->compressed_block;
-	int buffer_size = BGZF_BLOCK_SIZE;
-	int input_length = block_length;
-	int compressed_length = 0;
-	int remaining;
-	uint32_t crc;
-
-	assert(block_length <= BGZF_BLOCK_SIZE); // guaranteed by the caller
-	memcpy(buffer, g_magic, BLOCK_HEADER_LENGTH); // the last two bytes are a place holder for the length of the block
-	while (1) { // loop to retry for blocks that do not compress enough
-		int status;
-		z_stream zs;
-		zs.zalloc = NULL;
-		zs.zfree = NULL;
-		zs.next_in = fp->uncompressed_block;
-		zs.avail_in = input_length;
-		zs.next_out = (void*)&buffer[BLOCK_HEADER_LENGTH];
-		zs.avail_out = buffer_size - BLOCK_HEADER_LENGTH - BLOCK_FOOTER_LENGTH;
-		status = deflateInit2(&zs, fp->compress_level, Z_DEFLATED, -15, 8, Z_DEFAULT_STRATEGY); // -15 to disable zlib header/footer
-		if (status != Z_OK) {
-			fp->errcode |= BGZF_ERR_ZLIB;
-			return -1;
-		}
-		status = deflate(&zs, Z_FINISH);
-		if (status != Z_STREAM_END) { // not compressed enough
-			deflateEnd(&zs); // reset the stream
-			if (status == Z_OK) { // reduce the size and recompress
-				input_length -= 1024;
-				assert(input_length > 0); // logically, this should not happen
-				continue;
-			}
-			fp->errcode |= BGZF_ERR_ZLIB;
-			return -1;
-		}
-		if (deflateEnd(&zs) != Z_OK) {
-			fp->errcode |= BGZF_ERR_ZLIB;
-			return -1;
-		}
-		compressed_length = zs.total_out;
-		compressed_length += BLOCK_HEADER_LENGTH + BLOCK_FOOTER_LENGTH;
-		assert(compressed_length <= BGZF_BLOCK_SIZE);
-		break;
+	int comp_size = BGZF_MAX_BLOCK_SIZE;
+	if (bgzf_compress(fp->compressed_block, &comp_size, fp->uncompressed_block, block_length, fp->compress_level) != 0) {
+		fp->errcode |= BGZF_ERR_ZLIB;
+		return -1;
 	}
-
-	assert(compressed_length > 0);
-	packInt16((uint8_t*)&buffer[16], compressed_length - 1); // write the compressed_length; -1 to fit 2 bytes
-	crc = crc32(0L, NULL, 0L);
-	crc = crc32(crc, fp->uncompressed_block, input_length);
-	packInt32((uint8_t*)&buffer[compressed_length-8], crc);
-	packInt32((uint8_t*)&buffer[compressed_length-4], input_length);
-
-	remaining = block_length - input_length;
-	if (remaining > 0) {
-		assert(remaining <= input_length);
-		memcpy(fp->uncompressed_block, ((char*)fp->uncompressed_block) + input_length, remaining);
-	}
-	fp->block_offset = remaining;
-	return compressed_length;
+	fp->block_offset = 0;
+	return comp_size;
 }
 
 // Inflate the block in fp->compressed_block into fp->uncompressed_block
@@ -245,10 +225,10 @@ static int inflate_block(BGZF* fp, int block_length)
 	z_stream zs;
 	zs.zalloc = NULL;
 	zs.zfree = NULL;
-	zs.next_in = ((unsigned char*)fp->compressed_block) + 18;
+	zs.next_in = fp->compressed_block + 18;
 	zs.avail_in = block_length - 16;
 	zs.next_out = fp->uncompressed_block;
-	zs.avail_out = BGZF_BLOCK_SIZE;
+	zs.avail_out = BGZF_MAX_BLOCK_SIZE;
 
 	if (inflateInit2(&zs, -15) != Z_OK) {
 		fp->errcode |= BGZF_ERR_ZLIB;
@@ -279,7 +259,7 @@ static void free_cache(BGZF *fp)
 {
 	khint_t k;
 	khash_t(cache) *h = (khash_t(cache)*)fp->cache;
-	if (fp->open_mode != 'r') return;
+	if (fp->is_write) return;
 	for (k = kh_begin(h); k < kh_end(h); ++k)
 		if (kh_exist(h, k)) free(kh_val(h, k).block);
 	kh_destroy(cache, h);
@@ -296,7 +276,7 @@ static int load_block_from_cache(BGZF *fp, int64_t block_address)
 	if (fp->block_length != 0) fp->block_offset = 0;
 	fp->block_address = block_address;
 	fp->block_length = p->size;
-	memcpy(fp->uncompressed_block, p->block, BGZF_BLOCK_SIZE);
+	memcpy(fp->uncompressed_block, p->block, BGZF_MAX_BLOCK_SIZE);
 	_bgzf_seek((_bgzf_file_t)fp->fp, p->end_offset, SEEK_SET);
 	return p->size;
 }
@@ -307,8 +287,8 @@ static void cache_block(BGZF *fp, int size)
 	khint_t k;
 	cache_t *p;
 	khash_t(cache) *h = (khash_t(cache)*)fp->cache;
-	if (BGZF_BLOCK_SIZE >= fp->cache_size) return;
-	if ((kh_size(h) + 1) * BGZF_BLOCK_SIZE > fp->cache_size) {
+	if (BGZF_MAX_BLOCK_SIZE >= fp->cache_size) return;
+	if ((kh_size(h) + 1) * BGZF_MAX_BLOCK_SIZE > fp->cache_size) {
 		/* A better way would be to remove the oldest block in the
 		 * cache, but here we remove a random one for simplicity. This
 		 * should not have a big impact on performance. */
@@ -324,8 +304,8 @@ static void cache_block(BGZF *fp, int size)
 	p = &kh_val(h, k);
 	p->size = fp->block_length;
 	p->end_offset = fp->block_address + size;
-	p->block = malloc(BGZF_BLOCK_SIZE);
-	memcpy(kh_val(h, k).block, fp->uncompressed_block, BGZF_BLOCK_SIZE);
+	p->block = malloc(BGZF_MAX_BLOCK_SIZE);
+	memcpy(kh_val(h, k).block, fp->uncompressed_block, BGZF_MAX_BLOCK_SIZE);
 }
 #else
 static void free_cache(BGZF *fp) {}
@@ -339,7 +319,7 @@ int bgzf_read_block(BGZF *fp)
 	int count, size = 0, block_length, remaining;
 	int64_t block_address;
 	block_address = _bgzf_tell((_bgzf_file_t)fp->fp);
-	if (load_block_from_cache(fp, block_address)) return 0;
+	if (fp->cache_size && load_block_from_cache(fp, block_address)) return 0;
 	count = _bgzf_read(fp->fp, header, sizeof(header));
 	if (count == 0) { // no data read
 		fp->block_length = 0;
@@ -373,7 +353,7 @@ ssize_t bgzf_read(BGZF *fp, void *data, ssize_t length)
 	ssize_t bytes_read = 0;
 	uint8_t *output = data;
 	if (length <= 0) return 0;
-	assert(fp->open_mode == 'r');
+	assert(fp->is_write == 0);
 	while (bytes_read < length) {
 		int copy_length, available = fp->block_length - fp->block_offset;
 		uint8_t *buffer;
@@ -396,9 +376,169 @@ ssize_t bgzf_read(BGZF *fp, void *data, ssize_t length)
 	return bytes_read;
 }
 
+#if 0
+/***** BEGIN: multi-threading *****/
+
+typedef struct {
+	BGZF *fp;
+	struct mtaux_t *mt;
+	void *buf;
+	int i, errcode, toproc;
+} worker_t;
+
+typedef struct mtaux_t {
+	int n_threads, n_blks, curr, done;
+	volatile int proc_cnt;
+	void **blk;
+	int *len;
+	worker_t *w;
+	pthread_t *tid;
+	pthread_mutex_t lock;
+	pthread_cond_t cv;
+} mtaux_t;
+
+static int worker_aux(worker_t *w)
+{
+	int i, tmp, stop = 0;
+	// wait for condition: to process or all done
+	pthread_mutex_lock(&w->mt->lock);
+	while (!w->toproc && !w->mt->done)
+		pthread_cond_wait(&w->mt->cv, &w->mt->lock);
+	if (w->mt->done) stop = 1;
+	w->toproc = 0;
+	pthread_mutex_unlock(&w->mt->lock);
+	if (stop) return 1; // to quit the thread
+	w->errcode = 0;
+	for (i = w->i; i < w->mt->curr; i += w->mt->n_threads) {
+		int clen = BGZF_MAX_BLOCK_SIZE;
+		if (bgzf_compress(w->buf, &clen, w->mt->blk[i], w->mt->len[i], w->fp->compress_level) != 0)
+			w->errcode |= BGZF_ERR_ZLIB;
+		memcpy(w->mt->blk[i], w->buf, clen);
+		w->mt->len[i] = clen;
+	}
+	tmp = __sync_fetch_and_add(&w->mt->proc_cnt, 1);
+	return 0;
+}
+
+static void *mt_worker(void *data)
+{
+	while (worker_aux(data) == 0);
+	return 0;
+}
+
+int bgzf_mt(BGZF *fp, int n_threads, int n_sub_blks)
+{
+	int i;
+	mtaux_t *mt;
+	pthread_attr_t attr;
+	if (!fp->is_write || fp->mt || n_threads <= 1) return -1;
+	mt = calloc(1, sizeof(mtaux_t));
+	mt->n_threads = n_threads;
+	mt->n_blks = n_threads * n_sub_blks;
+	mt->len = calloc(mt->n_blks, sizeof(int));
+	mt->blk = calloc(mt->n_blks, sizeof(void*));
+	for (i = 0; i < mt->n_blks; ++i)
+		mt->blk[i] = malloc(BGZF_MAX_BLOCK_SIZE);
+	mt->tid = calloc(mt->n_threads, sizeof(pthread_t)); // tid[0] is not used, as the worker 0 is launched by the master
+	mt->w = calloc(mt->n_threads, sizeof(worker_t));
+	for (i = 0; i < mt->n_threads; ++i) {
+		mt->w[i].i = i;
+		mt->w[i].mt = mt;
+		mt->w[i].fp = fp;
+		mt->w[i].buf = malloc(BGZF_MAX_BLOCK_SIZE);
+	}
+	pthread_attr_init(&attr);
+	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
+	pthread_mutex_init(&mt->lock, 0);
+	pthread_cond_init(&mt->cv, 0);
+	for (i = 1; i < mt->n_threads; ++i) // worker 0 is effectively launched by the master thread
+		pthread_create(&mt->tid[i], &attr, mt_worker, &mt->w[i]);
+	fp->mt = mt;
+	return 0;
+}
+
+static void mt_destroy(mtaux_t *mt)
+{
+	int i;
+	// signal all workers to quit
+	pthread_mutex_lock(&mt->lock);
+	mt->done = 1; mt->proc_cnt = 0;
+	pthread_cond_broadcast(&mt->cv);
+	pthread_mutex_unlock(&mt->lock);
+	for (i = 1; i < mt->n_threads; ++i) pthread_join(mt->tid[i], 0); // worker 0 is effectively launched by the master thread
+	// free other data allocated on heap
+	for (i = 0; i < mt->n_blks; ++i) free(mt->blk[i]);
+	for (i = 0; i < mt->n_threads; ++i) free(mt->w[i].buf);
+	free(mt->blk); free(mt->len); free(mt->w); free(mt->tid);
+	pthread_cond_destroy(&mt->cv);
+	pthread_mutex_destroy(&mt->lock);
+	free(mt);
+}
+
+static void mt_queue(BGZF *fp)
+{
+	mtaux_t *mt = (mtaux_t*)fp->mt;
+	assert(mt->curr < mt->n_blks); // guaranteed by the caller
+	memcpy(mt->blk[mt->curr], fp->uncompressed_block, fp->block_offset);
+	mt->len[mt->curr] = fp->block_offset;
+	fp->block_offset = 0;
+	++mt->curr;
+}
+
+static int mt_flush(BGZF *fp)
+{
+	int i;
+	mtaux_t *mt = (mtaux_t*)fp->mt;
+	if (fp->block_offset) mt_queue(fp); // guaranteed that assertion does not fail
+	// signal all the workers to compress
+	pthread_mutex_lock(&mt->lock);
+	for (i = 0; i < mt->n_threads; ++i) mt->w[i].toproc = 1;
+	mt->proc_cnt = 0;
+	pthread_cond_broadcast(&mt->cv);
+	pthread_mutex_unlock(&mt->lock);
+	// worker 0 is doing things here
+	worker_aux(&mt->w[0]);
+	// wait for all the threads to complete
+	while (mt->proc_cnt < mt->n_threads);
+	// dump data to disk
+	for (i = 0; i < mt->n_threads; ++i) fp->errcode |= mt->w[i].errcode;
+	for (i = 0; i < mt->curr; ++i)
+		if (fwrite(mt->blk[i], 1, mt->len[i], fp->fp) != mt->len[i])
+			fp->errcode |= BGZF_ERR_IO;
+	mt->curr = 0;
+	return 0;
+}
+
+static int mt_lazy_flush(BGZF *fp)
+{
+	mtaux_t *mt = (mtaux_t*)fp->mt;
+	if (fp->block_offset) mt_queue(fp);
+	if (mt->curr == mt->n_blks)
+		return mt_flush(fp);
+	return -1;
+}
+
+static ssize_t mt_write(BGZF *fp, const void *data, ssize_t length)
+{
+	const uint8_t *input = data;
+	ssize_t rest = length;
+	while (rest) {
+		int copy_length = BGZF_BLOCK_SIZE - fp->block_offset < rest? BGZF_BLOCK_SIZE - fp->block_offset : rest;
+		memcpy(fp->uncompressed_block + fp->block_offset, input, copy_length);
+		fp->block_offset += copy_length; input += copy_length; rest -= copy_length;
+		if (fp->block_offset == BGZF_BLOCK_SIZE) mt_lazy_flush(fp);
+	}
+	return length - rest;
+}
+
+/***** END: multi-threading *****/
+#endif
+
 int bgzf_flush(BGZF *fp)
 {
-	assert(fp->open_mode == 'w');
+	if (!fp->is_write) return 0;
+	// disable mt
+        // if (fp->mt) return mt_flush(fp);
 	while (fp->block_offset > 0) {
 		int block_length;
 		block_length = deflate_block(fp, fp->block_offset);
@@ -414,18 +554,22 @@ int bgzf_flush(BGZF *fp)
 
 int bgzf_flush_try(BGZF *fp, ssize_t size)
 {
-	if (fp->block_offset + size > BGZF_BLOCK_SIZE)
-		return bgzf_flush(fp);
+	if (fp->block_offset + size > BGZF_BLOCK_SIZE) {
+          // disable mt
+          // if (fp->mt) return mt_lazy_flush(fp);
+          // else
+          return bgzf_flush(fp);
+	}
 	return -1;
 }
 
 ssize_t bgzf_write(BGZF *fp, const void *data, ssize_t length)
 {
 	const uint8_t *input = data;
-	int block_length = BGZF_BLOCK_SIZE, bytes_written;
-	assert(fp->open_mode == 'w');
-	input = data;
-	bytes_written = 0;
+	int block_length = BGZF_BLOCK_SIZE, bytes_written = 0;
+	assert(fp->is_write);
+	// disable mt
+        // if (fp->mt) return mt_write(fp, data, length);
 	while (bytes_written < length) {
 		uint8_t* buffer = fp->uncompressed_block;
 		int copy_length = block_length - fp->block_offset < length - bytes_written? block_length - fp->block_offset : length - bytes_written;
@@ -442,16 +586,20 @@ int bgzf_close(BGZF* fp)
 {
 	int ret, count, block_length;
 	if (fp == 0) return -1;
-	if (fp->open_mode == 'w') {
+	if (fp->is_write) {
 		if (bgzf_flush(fp) != 0) return -1;
+		fp->compress_level = -1;
 		block_length = deflate_block(fp, 0); // write an empty block
 		count = fwrite(fp->compressed_block, 1, block_length, fp->fp);
+                UNUSED(count);
 		if (fflush(fp->fp) != 0) {
 			fp->errcode |= BGZF_ERR_IO;
 			return -1;
 		}
+		// disable mt
+                // if (fp->mt) mt_destroy(fp->mt);
 	}
-	ret = fp->open_mode == 'w'? fclose(fp->fp) : _bgzf_close(fp->fp);
+	ret = fp->is_write? fclose(fp->fp) : _bgzf_close(fp->fp);
 	if (ret != 0) return -1;
 	free(fp->uncompressed_block);
 	free(fp->compressed_block);
@@ -472,8 +620,7 @@ int bgzf_check_EOF(BGZF *fp)
 	off_t offset;
 	offset = _bgzf_tell((_bgzf_file_t)fp->fp);
 	if (_bgzf_seek(fp->fp, -28, SEEK_END) < 0) return 0;
-	int n = _bgzf_read(fp->fp, buf, 28);
-    UNUSED(n);
+	_bgzf_read(fp->fp, buf, 28);
 	_bgzf_seek(fp->fp, offset, SEEK_SET);
 	return (memcmp(magic, buf, 28) == 0)? 1 : 0;
 }
@@ -483,7 +630,7 @@ int64_t bgzf_seek(BGZF* fp, int64_t pos, int where)
 	int block_offset;
 	int64_t block_address;
 
-	if (fp->open_mode != 'r' || where != SEEK_SET) {
+	if (fp->is_write || where != SEEK_SET) {
 		fp->errcode |= BGZF_ERR_MISUSE;
 		return -1;
 	}
