@@ -40,7 +40,7 @@
 #ifndef BZ_NO_STDIO
 void BZ2_bz__AssertH__fail ( int errcode )
 {
-   fprintf(stderr, 
+   REprintf( 
       "\n\nbzip2/libbzip2: internal error number %d.\n"
       "This is a bug in bzip2/libbzip2, %s.\n"
       "Please report it to me at: jseward@bzip.org.  If this happened\n"
@@ -54,7 +54,7 @@ void BZ2_bz__AssertH__fail ( int errcode )
    );
 
    if (errcode == 1007) {
-   fprintf(stderr,
+   REprintf(
       "\n*** A special note about internal error number 1007 ***\n"
       "\n"
       "Experience suggests that a common cause of i.e. 1007\n"
